import json
from pathlib import Path
from typing import Any, Iterator, cast

import pytest
from pgstacrs import Client
from pypgstac.db import PgstacDB
from pypgstac.migrate import Migrate
from pytest import Config, FixtureRequest, Parser
from pytest_postgresql import factories
from pytest_postgresql.executor import PostgreSQLExecutor
from pytest_postgresql.janitor import DatabaseJanitor


def migrate(
    toversion: str, host: str, port: int, user: str, dbname: str, password: str
) -> None:
    pgstac_db = PgstacDB(f"postgresql://{user}:{password}@{host}:{port}/{dbname}")
    Migrate(pgstac_db).run_migration(toversion)


pgstac_proc = factories.postgresql_proc()


@pytest.fixture(scope="session", params=["0.8.5", "0.9.1"])
def pgstac(
    pgstac_proc: PostgreSQLExecutor, request: FixtureRequest
) -> Iterator[PostgreSQLExecutor]:
    migrate(
        request.param,
        pgstac_proc.host,
        pgstac_proc.port,
        pgstac_proc.user,
        pgstac_proc.template_dbname,
        pgstac_proc.password,
    )
    yield pgstac_proc


@pytest.fixture
def database_janitor(pgstac: PostgreSQLExecutor) -> Iterator[DatabaseJanitor]:
    with DatabaseJanitor(
        user=pgstac.user,
        host=pgstac.host,
        port=pgstac.port,
        version=pgstac.version,
        password=pgstac.password,
        dbname="pypgstac_test",
        template_dbname=pgstac.template_dbname,
    ) as database_janitor:
        yield database_janitor


@pytest.fixture
async def client(database_janitor: DatabaseJanitor) -> Client:
    return await Client.open(
        f"user={database_janitor.user} host={database_janitor.host} port={database_janitor.port} dbname={database_janitor.dbname} password={database_janitor.password}"
    )


@pytest.fixture
def collection(examples_path: Path) -> dict[str, Any]:
    with open(examples_path / "collection.json") as f:
        return cast(dict[str, Any], json.load(f))


@pytest.fixture
def item(examples_path: Path) -> dict[str, Any]:
    with open(examples_path / "simple-item.json") as f:
        return cast(dict[str, Any], json.load(f))


@pytest.fixture
def examples_path() -> Path:
    return Path(__file__).parents[1] / "spec-examples" / "v1.0.0"


def pytest_addoption(parser: Parser) -> None:
    parser.addoption(
        "--external",
        action="store_true",
        default=False,
        help="run tests that require an external database via docker compose",
    )


def pytest_collection_modifyitems(config: Config, items: Any) -> None:
    if config.getoption("--external"):
        return
    skip_external = pytest.mark.skip(reason="need --external option to run")
    for item in items:
        if "external" in item.keywords:
            item.add_marker(skip_external)



from typing import Any

import pytest
from pgstacrs import Client
from pytest_postgresql import factories
from pytest_postgresql.executor import PostgreSQLExecutor

pytestmark = pytest.mark.external

pgstac_in_docker = factories.postgresql_noproc(
    host="localhost", port=5432, user="username", password="password", dbname="pgstac"
)


async def test_external(
    pgstac_in_docker: PostgreSQLExecutor,
    collection: dict[str, Any],
    item: dict[str, Any],
) -> None:
    client = await Client.open(
        f"postgresql://{pgstac_in_docker.user}:{pgstac_in_docker.password}@{pgstac_in_docker.host}:{pgstac_in_docker.port}/{pgstac_in_docker.dbname}"
    )
    await client.upsert_collection(collection)
    await client.create_item(item)
    await client.delete_item("20201211_223832_CS2")



from typing import Any

import pytest
from pgstacrs import Client, PgstacError


async def test_get_and_create_collection(
    client: Client, collection: dict[str, Any]
) -> None:
    assert await client.get_collection("simple-collection") is None
    await client.create_collection(collection)
    assert await client.get_collection("simple-collection") is not None
    with pytest.raises(PgstacError, match="already exists"):  # type: ignore
        await client.create_collection(collection)


async def test_update_collection(client: Client, collection: dict[str, Any]) -> None:
    with pytest.raises(PgstacError, match="no rows"):  # type: ignore
        await client.update_collection(collection)
    await client.create_collection(collection)
    db_collection = await client.get_collection("simple-collection")
    assert db_collection
    assert db_collection["description"] != "a new description"
    collection["description"] = "a new description"
    await client.update_collection(collection)
    db_collection = await client.get_collection("simple-collection")
    assert db_collection
    assert db_collection["description"] == "a new description"


async def test_upsert_collection(client: Client, collection: dict[str, Any]) -> None:
    await client.upsert_collection(collection)
    db_collection = await client.get_collection("simple-collection")
    assert db_collection
    assert db_collection["description"] != "a new description"
    collection["description"] = "a new description"
    await client.upsert_collection(collection)
    db_collection = await client.get_collection("simple-collection")
    assert db_collection
    assert db_collection["description"] == "a new description"


async def test_delete_collection(client: Client, collection: dict[str, Any]) -> None:
    with pytest.raises(PgstacError, match="no rows"):  # type: ignore
        await client.delete_collection("simple-collection")
    await client.create_collection(collection)
    await client.delete_collection("simple-collection")
    assert await client.get_collection("simple-id") is None


async def test_all_collections(client: Client, collection: dict[str, Any]) -> None:
    assert len(await client.all_collections()) == 0
    await client.create_collection(collection)
    assert len(await client.all_collections()) == 1
    collection["id"] = "just-as-simple-collection"
    await client.create_collection(collection)
    assert len(await client.all_collections()) == 2
    await client.delete_collection("simple-collection")
    assert len(await client.all_collections()) == 1



from typing import Any

import pytest
from pgstacrs import Client, PgstacError


async def test_get_and_create_item(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    assert await client.get_item("20201211_223832_CS2") is None
    await client.create_collection(collection)
    assert await client.get_item("20201211_223832_CS2") is None
    await client.create_item(item)
    assert await client.get_item("20201211_223832_CS2") is not None
    assert await client.get_item("20201211_223832_CS2", "simple-collection") is not None

    item["collection"] = "does-not-exist"
    with pytest.raises(PgstacError, match="does-not-exist"):  # type: ignore
        await client.create_item(item)


async def test_update_item(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    with pytest.raises(PgstacError, match="query returned no rows"):  # type: ignore
        await client.update_item(item)
    await client.create_collection(collection)
    with pytest.raises(PgstacError, match="query returned no rows"):  # type: ignore
        await client.update_item(item)
    await client.create_item(item)
    item["properties"]["foo"] = "bar"
    await client.update_item(item)
    db_item = await client.get_item("20201211_223832_CS2")
    assert db_item
    assert db_item["properties"]["foo"] == "bar"


async def test_upsert_item(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    with pytest.raises(PgstacError, match="simple-collection"):  # type: ignore
        await client.upsert_item(item)
    await client.create_collection(collection)
    await client.upsert_item(item)
    assert client.get_item("20201211_223832_CS2") is not None
    item["properties"]["foo"] = "bar"
    await client.upsert_item(item)
    # TODO ensure there's only one item
    db_item = await client.get_item("20201211_223832_CS2")
    assert db_item
    assert db_item["properties"]["foo"] == "bar"


async def test_create_items(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.create_items([item])
    assert await client.get_item("20201211_223832_CS2") is not None


async def test_upsert_items(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.upsert_items([item])
    assert await client.get_item("20201211_223832_CS2") is not None
    item["properties"]["foo"] = "bar"
    await client.upsert_items([item])
    db_item = await client.get_item("20201211_223832_CS2")
    assert db_item
    assert db_item["properties"]["foo"] == "bar"


async def test_delete_item(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    with pytest.raises(PgstacError, match="no rows"):  # type: ignore
        await client.delete_item("20201211_223832_CS2")
    with pytest.raises(PgstacError, match="simple-collection"):  # type: ignore
        await client.create_item(item)

    await client.create_collection(collection)
    await client.create_item(item)
    assert await client.get_item("20201211_223832_CS2") is not None
    await client.delete_item("20201211_223832_CS2")
    assert await client.get_item("20201211_223832_CS2") is None

    with pytest.raises(PgstacError, match="no rows"):  # type: ignore
        await client.delete_item("20201211_223832_CS2")

    await client.create_item(item)
    await client.delete_item("20201211_223832_CS2", "simple-collection")
    assert await client.get_item("20201211_223832_CS2") is None


async def test_update_collection_extents(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    collection["extent"]["spatial"]["bbox"] = [[-180, -90, 180, 90]]
    await client.create_collection(collection)
    await client.create_item(item)
    db_collection = await client.get_collection("simple-collection")
    assert db_collection
    assert db_collection["extent"]["spatial"]["bbox"] == [[-180, -90, 180, 90]]
    await client.update_collection_extents()
    db_collection = await client.get_collection("simple-collection")
    assert db_collection
    assert db_collection["extent"]["spatial"]["bbox"] != [[-180, -90, 180, 90]]



import copy
import urllib.parse
from typing import Any

import pytest
from pgstacrs import Client


async def test_empty_search(client: Client) -> None:
    search = await client.search()
    version = await client.get_version()
    if version.startswith("0.9"):
        assert search == {
            "features": [],
            "links": [
                {"href": ".", "rel": "root", "type": "application/json"},
                {"href": "./search", "rel": "self", "type": "application/json"},
            ],
            "numberReturned": 0,
            "type": "FeatureCollection",
        }
    elif version.startswith("0.8"):
        assert search == {
            "type": "FeatureCollection",
            "context": {"limit": 10, "returned": 0},
            "features": [],
        }


async def test_base_url(client: Client) -> None:
    await client.set_setting("base_url", "http://pgstacrs.test")
    search = await client.search()
    version = await client.get_version()
    if version.startswith("0.9"):
        assert search == {
            "features": [],
            "links": [
                {
                    "href": "http://pgstacrs.test/",
                    "rel": "root",
                    "type": "application/json",
                },
                {
                    "href": "http://pgstacrs.test/search",
                    "rel": "self",
                    "type": "application/json",
                },
            ],
            "numberReturned": 0,
            "type": "FeatureCollection",
        }


async def test_search(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.create_item(item)
    feature_collection = await client.search()
    assert len(feature_collection["features"]) == 1


async def test_search_fields(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.create_item(item)

    feature_collection = await client.search(include="id")
    item = feature_collection["features"][0]
    assert item == {"id": "20201211_223832_CS2", "collection": "simple-collection"}

    feature_collection = await client.search(exclude="id")
    item = feature_collection["features"][0]
    assert "id" not in item


@pytest.mark.skip("I'm not sure query is implemented properly in pgstac?")
async def test_search_query(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    item["properties"]["foo"] = "bar"
    await client.create_item(item)

    feature_collection = await client.search(query={"query": {"foo": {"eq": "bar"}}})
    assert len(feature_collection["features"]) == 1

    feature_collection = await client.search(query={"query": {"foo": {"eq": "baz"}}})
    assert len(feature_collection["features"]) == 0


async def test_bbox(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.create_item(item)

    feature_collection = await client.search(bbox=[170, 0, 173, 2])
    assert len(feature_collection["features"]) == 1

    # Looks like my postgres doesn't like 3d bboxes
    # feature_collection = await client.search(bbox=[170, 0, -1000, 173, 2, 20000])
    # assert len(feature_collection["features"]) == 1

    feature_collection = await client.search(bbox=[0, 0, 1, 1])
    assert len(feature_collection["features"]) == 0


async def test_sortby(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    item_a = copy.deepcopy(item)
    item_a["id"] = "a"
    item_a["properties"]["foo"] = "a"
    item_a["properties"]["bar"] = 0
    item_b = copy.deepcopy(item)
    item_b["id"] = "b"
    item_b["properties"]["foo"] = "b"
    item_b["properties"]["bar"] = 1
    item_c = copy.deepcopy(item)
    item_c["id"] = "c"
    item_c["properties"]["foo"] = "c"
    item_c["properties"]["bar"] = 1
    await client.create_items([item_a, item_b, item_c])

    feature_collection = await client.search(sortby="+foo")
    assert feature_collection["features"][0]["id"] == "a"
    assert feature_collection["features"][1]["id"] == "b"

    feature_collection = await client.search(sortby="foo")
    assert feature_collection["features"][0]["id"] == "a"
    assert feature_collection["features"][1]["id"] == "b"

    feature_collection = await client.search(sortby="-foo")
    assert feature_collection["features"][0]["id"] == "c"
    assert feature_collection["features"][1]["id"] == "b"

    feature_collection = await client.search(sortby=["-bar", "+foo"])
    assert feature_collection["features"][0]["id"] == "b"
    assert feature_collection["features"][1]["id"] == "c"

    feature_collection = await client.search(sortby=["-bar", "-foo"])
    assert feature_collection["features"][0]["id"] == "c"
    assert feature_collection["features"][1]["id"] == "b"


async def test_filter(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    item["properties"]["foo"] = "bar"
    await client.create_item(item)

    feature_collection = await client.search(filter="foo = 'bar'")
    assert len(feature_collection["features"]) == 1
    feature_collection = await client.search(filter="foo != 'bar'")
    assert len(feature_collection["features"]) == 0

    feature_collection = await client.search(
        filter={"op": "=", "args": [{"property": "foo"}, "bar"]}
    )
    assert len(feature_collection["features"]) == 1
    feature_collection = await client.search(
        filter={"op": "!=", "args": [{"property": "foo"}, "bar"]}
    )
    assert len(feature_collection["features"]) == 0


async def test_intersects(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.create_item(item)

    feature_collection = await client.search(
        intersects={"type": "Point", "coordinates": [0, 0]}
    )
    assert len(feature_collection["features"]) == 0

    feature_collection = await client.search(
        intersects={"type": "Point", "coordinates": [172.92, 1.35]}
    )
    assert len(feature_collection["features"]) == 1

    feature_collection = await client.search(
        intersects='{"type": "Point", "coordinates": [172.92, 1.35]}'
    )
    assert len(feature_collection["features"]) == 1


async def test_ids(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.create_item(item)

    feature_collection = await client.search(ids="not-an-id")
    assert len(feature_collection["features"]) == 0

    feature_collection = await client.search(ids="20201211_223832_CS2")
    assert len(feature_collection["features"]) == 1

    feature_collection = await client.search(ids=["20201211_223832_CS2"])
    assert len(feature_collection["features"]) == 1


async def test_collections(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    await client.create_collection(collection)
    await client.create_item(item)

    feature_collection = await client.search(collections="not-an-id")
    assert len(feature_collection["features"]) == 0

    feature_collection = await client.search(collections="simple-collection")
    assert len(feature_collection["features"]) == 1

    feature_collection = await client.search(collections=["simple-collection"])
    assert len(feature_collection["features"]) == 1


async def test_pagination(
    client: Client, collection: dict[str, Any], item: dict[str, Any]
) -> None:
    version = await client.get_version()

    await client.create_collection(collection)
    item["id"] = "first-item"
    await client.create_item(item)
    second_item = copy.deepcopy(item)
    second_item["id"] = "second-item"
    await client.create_item(second_item)

    feature_collection = await client.search(limit=1, sortby="id")

    if version.startswith("0.9"):
        next_link = next(
            (link for link in feature_collection["links"] if link["rel"] == "next")
        )
        url = urllib.parse.urlparse(next_link["href"])
        token = url.query.split("=")[1]
    elif version.startswith("0.8"):
        token = "next:" + feature_collection["next"]

    feature_collection = await client.search(limit=1, sortby="id", token=token)
    assert feature_collection["features"][0]["id"] == "second-item"



from typing import Any

class Client:
    """A pgstac client."""

    @classmethod
    async def open(cls: type[Client], params: str) -> Client:
        """Opens a new client, backed by a connection pool.

        Args:
            params: The connection parameters, either in `postgresql://` or
                `host=localhost user=username` format.

        Returns:
            A pgstac client

        Examples:
            >>> from pgstacrs import Client
            >>> client = await Client.open("postgresql://username:password@localhost:5432/pgstac")
            >>> client = await Client.open("user=username password=password dbname=pgstac")
        """

    async def search(
        self,
        *,
        intersects: str | dict[str, Any] | None = None,
        ids: str | list[str] | None = None,
        collections: str | list[str] | None = None,
        limit: int | None = None,
        bbox: list[float] | None = None,
        datetime: str | None = None,
        include: str | list[str] | None = None,
        exclude: str | list[str] | None = None,
        sortby: str | list[str] | None = None,
        filter: str | dict[str, Any] | None = None,
        query: dict[str, Any] | None = None,
        **kwargs: str,
    ) -> dict[str, Any]:
        """
        Searches the database with STAC API item search.

        Args:
            collections: Array of one or more Collection IDs that
                each matching Item must be in.
            ids: Array of Item ids to return.
            intersects: Searches items by performing intersection between their
                geometry and provided GeoJSON geometry.
            bbox: Requested bounding box.
            datetime: Single date+time, or a range (`/` separator), formatted to
                RFC 3339, section 5.6.  Use double dots .. for open date ranges.
            include: Fields to include in the response (see [the extension
                docs](https://github.com/stac-api-extensions/fields?tab=readme-ov-file#includeexclude-semantics))
                for more on the semantics).
            exclude: Fields to exclude from the response (see [the extension
                docs](https://github.com/stac-api-extensions/fields?tab=readme-ov-file#includeexclude-semantics))
                for more on the semantics).
            sortby: Fields by which to sort results (use `-field` to sort descending).
            filter: CQL2 filter expression. Strings will be interpreted as
                cql2-text, dictionaries as cql2-json.
            query: Additional filtering based on properties.
                It is recommended to use filter instead, if possible.
            limit: The page size returned from the server.
            kwargs: Any additional arguments to pass down into the search, e.g a pagination token
        """

    async def print_config(self) -> None:
        """Prints the postgresql configuration.

        Redacts the password
        """

    async def set_setting(self, key: str, value: str) -> None:
        """Sets a pgstac setting.

        Args:
            key: The setting name, e.g. `base_url`
            value: The setting value, e.g. `http://pgstacrs.test`
        """

    async def get_version(self) -> str:
        """Returns the pgstac version.

        Returns:
            The pgstac version as a string
        """

    async def get_collection(self, id: str) -> dict[str, Any] | None:
        """Returns a collection by id, or none if one does not exist.

        Args:
            id: The collection id

        Returns:
            A STAC collection, or None
        """

    async def create_collection(self, collection: dict[str, Any]) -> None:
        """Creates a new collection.

        Args:
            collection: The collection

        Raises:
            PgstacError: If the collection already exists.
        """

    async def update_collection(self, collection: dict[str, Any]) -> None:
        """Updates a collection.

        Args:
            collection: The collection

        Raises:
            PgstacError: If the collection does not exist.
        """

    async def update_collection_extents(self) -> None:
        """Updates all collection extents."""

    async def upsert_collection(self, collection: dict[str, Any]) -> None:
        """Upserts a collection.

        Args:
            collection: The collection
        """

    async def delete_collection(self, id: str) -> None:
        """Deletes a collection by id.

        Args:
            id: The collection id
        """

    async def all_collections(self) -> list[dict[str, Any]]:
        """Returns all collections.

        Returns:
            All collections in the database
        """

    async def get_item(
        self, id: str, collection_id: str | None = None
    ) -> dict[str, Any] | None:
        """Returns an item by id.

        Args:
            id: The item id
            collection_id: The optional collection id

        Returns:
            The item, or None if the item does not exist
        """

    async def create_item(self, item: dict[str, Any]) -> None:
        """Creates an item.

        Args:
            item: The item

        Raises:
            PgstacError: If the item's collection does not exist. The collection
                is determined by the `collection` attribute of the item.
        """

    async def create_items(self, items: list[dict[str, Any]]) -> None:
        """Creates many items.

        Args:
            items: The items

        Raises:
            PgstacError: If the items' collection(s) does not exist.
        """

    async def update_item(self, item: dict[str, Any]) -> None:
        """Updates an item.

        Args:
            item: The item

        Raises:
            PgstacError: If the item does not exist
        """

    async def upsert_item(self, item: dict[str, Any]) -> None:
        """Upserts an item.

        Args:
            item: The item

        Raises:
            PgstacError: If the item's collection does not exist.
        """

    async def upsert_items(self, items: list[dict[str, Any]]) -> None:
        """Upserts many items.

        Args:
            items: The items

        Raises:
            PgstacError: If the items' collection(s) does not exist.
        """

    async def delete_item(self, id: str, collection_id: str | None = None) -> None:
        """Deletes an item by id.

        Args:
            id: The item id
            collection_id: The optional collection id

        Raises:
            PgstacError: If the item cannot be found
        """

class PgstacError:
    """An exception returned from pgstac"""

class StacError:
    """Something doesn't match the STAC specification"""
